// makerom.c -- hex to binary converter -- by Bruce Tomlin

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#define version "Makerom version 1.2"
typedef unsigned char bool;		// i guess bool is a c++ thing
const bool FALSE = 0;
const bool TRUE = 1;
#define IHEX_SIZE	32		// max number of data bytes per line in intel hex format

unsigned char	data[65536];
int				loaddr;			// lowest address encountered in data
int 			hiaddr;			// highest address encountered in data
int				imgsize;		// image size in bytes
int				baseaddr;		// image base address
int				loadaddr;		// image load address
int				xferaddr;		// transfer address
int				fill;			// undefined memory fill value
char			outfmt;			// 'I' = intel hex, 'S' = motorola s-record, else binary
char			*fname;			// input file name
const char		*progname;		// pointer to argv[0]


int ishex(char c)
{
	c = toupper(c);
	return isdigit(c) || ('A' <= c && c <= 'F');
}


int hexdigit(char c)
{
	c = toupper(c);
	if (isdigit(c))		return c - '0';
	if ('A' <= c && c <= 'F')	return c - 'A' + 10;
	return -1;
}


void setloaddr(unsigned int addr)
{
	if (loaddr == -1 || addr < loaddr)
		loaddr = addr;
}


void sethiaddr(unsigned int addr)
{
	if (hiaddr == -1 || addr > hiaddr)
		hiaddr = addr;
}


// Intel hex format:
//
// :aabbbbccdddd...ddee
//
// aa    = record data length (the number of dd bytes)
// bbbb  = address for this record
// cc    = record type
//       00 = data (data is in the dd bytes)
//       01 = end of file (bbbb is transfer address)
//       02 = extended segment address record
//            dddd (two bytes) represents the segment address
//       03 = Start segment address record
//            dddd (two bytes) represents the segment of the transfer address
//       04 = extended linear address record
//            dddd (two bytes) represents the high address word
//       05 = Start linear address record
//            dddd (two bytes) represents the high word of the transfer address
// dd... = data bytes if record type needs it
// ee    = checksum byte: add all bytes aa through dd
//                        and subtract from 256 (2's complement negate)
//
// Note that record types 02 thru 05 are not used, because this utility is
// currently intended only for 16-bit address spaces.


void write_ihex(int addr, unsigned char *buf, int len, int rectype)
{
	int i,chksum;

	chksum = len + (addr >> 8) + addr + rectype;
	printf(":%.2X%.4X%.2X",len,addr & 0xFFFF,rectype);
	for (i=0; i<len; i++)
	{
		printf("%.2X",buf[i]);
		chksum = chksum + buf[i];
	}
	printf("%.2X\n",(-chksum) & 0xFF);
}


void read_ihex(char *line)
{
	unsigned char	buf[256];
	int				n,len;
	int				i;
	char			*err;
	int				hi,lo;
	int				checksum;
	int				addr;
	int				typ;

	err = NULL;

	// make sure line is long enough and has an even number of hex digits
	len = strlen(line);
//	if (line[1] != ':')
//		err = "Intel hex data must start with a colon";
//	else
	if (!(len & 1) || len < 11)
		err = "Invalid length of Intel hex data line";
	else
	{
		// convert line from hex to binary
		n = len/2;
		for (i = 0; i < n && err == NULL; i++)
		{
			hi = hexdigit(line[i*2+1]);
			lo = hexdigit(line[i*2+2]);
			if (hi < 0 || lo < 0)		// validate the hex digits
				err = "Non-hexadecimal character in Intel hex data line";
			else
				buf[i] = hi*16 + lo;
		}

		if (buf[0] != n-5)				// validate the length byte
			err = "Incorrect length byte in Intel hex data line";
		else
		{								// validate the checksum
			checksum = 0;
			for (i=0; i < n-1; i++)
				checksum = checksum + buf[i];
			checksum = (-checksum) & 0xFF;
			if (checksum != buf[i])
				err = "Invalid checksum in Intel hex data line";
		}
	}

	if (err)
		fprintf(stderr,"%s\nError: %s\n",line,err);
	else
	{
		addr = buf[1]*256 + buf[2];
		typ  = buf[3];
		len  = buf[0];
		// data = buf[4] .. buf[n-2]
//		printf("addr = %.4X, typ = %d\n",addr,typ);
		switch(typ)
		{
			case 0:
				if (len)	// ignore empty data records!
				{
					setloaddr(addr);			// update lowest and highest seen addresses
					sethiaddr(addr + len-1);
					for (i = 0; i < len; i++)	// copy data from line
						data[(addr + i) & 0xFFFF] = buf[i+4];
				}
				break;

			case 1:
				// need to translate address and send out new xfer address record
				break;
		}
	}
}


// Motorola S-record format:
//
// Sabbcc...ccdd...ddee
//
// a     = record type
//         0 starting record (optional)
//         1 data record with 16-bit address
//         2 data record with 24-bit address
//         3 data record with 32-bit address
//         4 symbol record (LSI extension) - S4<length><address><name>,<checksum>
//         5 number of data records in preceeding block
//         6 unused
//         7 ending record for S3 records
//         8 ending record for S2 records
//         9 ending record for S1 records
// bb    = record data length (from bb through ee)
// cc... = address for this record, 2 bytes for S1/S9, 3 bytes for S2/S8, 4 bytes for S3/S7
// dd... = data bytes if record type needs it
// ee    = checksum byte: add all bytes bb through dd
//                        and subtract from 255 (1's complement)
//
// Note that record types S2 thru S8 are not used, because this utility is
// currently intended only for 16-bit address spaces.


void write_srec(int addr, unsigned char *buf, int len, int rectype)
{
	int i,chksum;

	chksum = len+3 + (addr >> 8) + addr;
	printf("S%d%.2X%.4X", rectype, len+3, addr & 0xFFFF);
	for (i=0; i<len; i++)
	{
		printf("%.2X",buf[i]);
		chksum = chksum + buf[i];
	}
	printf("%.2X\n",~chksum & 0xFF);
}


void read_srec(char *line)
{
	unsigned char	buf[256];
	int				n,len;
	int				i;
	char			*err;
	int				hi,lo;
	int				checksum;
	int				addr;

	err = NULL;

	// make sure line is long enough and has an even number of hex digits
	len = strlen(line);
//	if (toupper(line[1]) != 'S')
//		err = "S-record data must start with an 'S'";
//	else
	if ((len & 1) || len < 10)
		err = "Invalid length of S-record data line";
	else
	{
		// convert line from hex to binary
		n = len/2 - 1;
		for (i = 0; i < n && err == NULL; i++)
		{
			hi = hexdigit(line[i*2+2]);
			lo = hexdigit(line[i*2+3]);
			if (hi < 0 || lo < 0)		// validate the hex digits
				err = "Non-hexadecimal character in S-record data line";
			else
				buf[i] = hi*16 + lo;
		}

		if (buf[0] != n-1)				// validate the length byte
			err = "Incorrect length byte in S-record data line";
		else
		{								// validate the checksum
			checksum = 0;
			for (i=0; i < n-1; i++)
				checksum = checksum + buf[i];
			checksum = ~checksum & 0xFF;
			if (checksum != buf[i])
				err = "Invalid checksum in S-record data line";
		}
	}

	if (err)
		fprintf(stderr,"%s\nError: %s\n",line,err);
	else
	{
		switch(line[1])
		{
			case '1':
			case '2':
			case '3':
				n = line[1] - '1';
				addr = buf[n+1]*256 + buf[n+2];
				len  = buf[0] - n - 3;

				// data = buf[n+3] .. buf[n+len+3-1]

				if (len)	// ignore empty data records!
				{
					setloaddr(addr);			// update lowest and highest seen addresses
					sethiaddr(addr + len-1);
					for (i = 0; i < len; i++)	// copy data from line
						data[(addr + i) & 0xFFFF] = buf[i+n+3];
				}
				break;

			case '7':
			case '8':
			case '9':
				n = -(line[1] - '9');
				addr = buf[n+1]*256 + buf[n+2];
				xferaddr = addr;
				break;
		}
	}
}


void doline(char *line)
{
	switch(toupper(line[0]))
	{
		case ':':	// intel hex record data
			read_ihex(line);
			break;

		case 'S':	// motorola s-record data
			read_srec(line);
			break;

		default:	// ignore anything else
			break;
	}
}


void usage(void)
{
	printf("%s\n",version);
	printf("Usage:\n");
	printf("    %s [options] [input file]\n",progname);
	printf("Options:\n");
	printf("    -h         output in Intel hex format\n");
	printf("    -s         output in Motorola S-record format\n");
	printf("    -b addr    base address of image (hex, default = automatic)\n");
	printf("    -l addr    load address of hex output image (hex, default = 0)\n");
	printf("    -k size    image size in kbytes (decimal, default = automatic)\n");
	printf("    -f value   fill value for undefined memory locations (hex, default = FF)\n");
	exit(1);
}


void getopts(int argc, char * const argv[])
{
	int bflag, ch;

	bflag = 0;
	while ((ch = getopt(argc, argv, "b:l:f:hsk:?")) != -1)
	{
		switch (ch)
		{
			case 'b':
				sscanf(optarg, "%x", &baseaddr);
				break;

			case 'l':
				sscanf(optarg, "%x", &loadaddr);
				break;

			case 'f':
				sscanf(optarg, "%x", &fill);
				break;

			case 'h':
				outfmt = 'I';
				break;

			case 's':
				outfmt = 'S';
				break;

			case 'k':
				imgsize = atoi(optarg) * 1024;
				break;

			case '?':
			default:
				usage();
		}
	}
	argc -= optind;
	argv += optind;

	// now argc is the number of remaining arguments
	// and argv[0] is the first remaining argument

	fname = NULL;

	if (argc > 1)
		usage();

	if (argc)
		fname = argv[0];

//	if (fname[0] == '?' && fname[1] == 0)
//		usage();
}


int main (int argc, char * const argv[])
{
	char	c;
	FILE	*f;
	char	line[256];
	char	*p;
	int		addr;
	int		n;

	progname = argv[0];

	loaddr   = -1;
	hiaddr   = -1;
	imgsize  = -1;
	baseaddr = -1;
	loadaddr =  0;
	xferaddr = -1;
	outfmt   = 'B';
	fill     = 0xFF;
/*
	addr = 0;
	xferaddr = 0;
	showxfer = FALSE;
	delay = 0;
	linelength = 32;
*/
	getopts(argc, argv);

	memset(data, fill, sizeof(data));

	f = NULL;
	if (fname)	f = fopen(fname, "r");
	if (!f)		f = stdin;

	// FIXME: do I need to add a binary read mode option?

	line[0] = 0;
	p = line;
	while ((c = fgetc(f)) != EOF)
	{
		switch(c)
		{
			case 0:
				break;

			case '\r':
			case '\n':
				doline(line);
				line[0] = 0;
				p = line;
				break;

			default:
				*p++ = c;
				*p = 0;
				break;
		}
	}
	if (line[0])
		doline(line);

//printf("baseaddr = %.4X, imgsize = %.4X, hiaddr = %.4X, loaddr = %.4X\n",
//		baseaddr, imgsize, hiaddr, loaddr);

	// handle automatic default image size
	if (baseaddr == -1)
		baseaddr = loaddr;
	if (baseaddr == -1)
		baseaddr = 0;
	if (imgsize == -1)
	{
		if (hiaddr == -1)	imgsize = 1024;
		else				imgsize = hiaddr - baseaddr + 1;
	}

//printf("baseaddr = %.4X, imgsize = %.4X, hiaddr = %.4X, loaddr = %.4X\n",
//		baseaddr, imgsize, hiaddr, loaddr);

	// now dump what we've read in
// dump hex data from baseaddr to baseaddr + imgsize - 1
// offseting address by: addr = addr - baseaddr + loadaddr;
	if (outfmt == 'I' || outfmt == 'S')
	{
		addr = baseaddr;
		while (addr - baseaddr < imgsize)
		{
			n = imgsize - (addr - baseaddr);
			if (n > IHEX_SIZE)
				n = IHEX_SIZE;

//			printf(">addr = %.4X, newaddr = %.4X, n = %d\n",addr,addr-baseaddr+loadaddr,n);
			switch(outfmt)
			{
				default:
				case 'I':
					write_ihex(addr - baseaddr + loadaddr, data+addr, n, 0);
					break;
				case 'S':
					write_srec(addr - baseaddr + loadaddr, data+addr, n, 1);
					break;
			}

			addr = addr + n;
		}
		// write xfer address here
/*
		if (xferaddr != -1)
		{
			switch(outfmt)
			{
				default:
				case 'I':
					write_ihex(xferaddr - baseaddr + loadaddr, data+addr, n, 1);
					break;
				case 'S':
					write_srec(xferaddr - baseaddr + loadaddr, data+addr, n, 9);
					break;
			}
		}
*/
	}
	// binary file write
	// end = baseaddr + imgsize - 1
	// len = (baseaddr + imgsize) - baseaddr
	else fwrite(data+baseaddr, 1, imgsize, stdout);

	if (fname)
		fclose(f);

	return 0;
}
